print("import: 'azure'")
import azure

print("import: 'azure.storage'")
import azure.storage

print("import: 'azure.storage._http'")
import azure.storage._http

print("import: 'azure.storage.blob'")
import azure.storage.blob

print("import: 'azure.storage.file'")
import azure.storage.file

print("import: 'azure.storage.queue'")
import azure.storage.queue

print("import: 'azure.storage.table'")
import azure.storage.table

